-- scope stuff
qualityoflife = qualityoflife or {};
local q = qualityoflife;

q.classicheals = q.classicheals or {};
local c = q.classicheals;

QOL_OPTIONS = QOL_OPTIONS or {};
QOL_OPTIONS["ClassicHeals"] = QOL_OPTIONS["ClassicHeals"] or {};
local o = QOL_OPTIONS["ClassicHeals"];

-- variables
local targetFrame;
local partyFrames = {};
local raidFrames_Compact = {};
local raidFrames_Pullout = {};

local OPT_FONTSIZE = "OPT_FONTSIZE";
local OPT_SHOWOVERHEAL = "OPT_SHOWOVERHEAL";

-- init
function c:Init()
	if not q.initFinished then
		q.Init();
	end

	o[OPT_FONTSIZE] = o[OPT_FONTSIZE] or 16;
	if o[OPT_SHOWOVERHEAL] == nil then
		o[OPT_SHOWOVERHEAL] = true;
	end

	c.initFinished = true;
	c:Println(q:GetText("Module loaded."));
end

-- command handling
local CMD_FONTSIZE = "fontsize";
local CMD_OVERHEAL = "overheal";

q.commandHandlers = q.commandHandlers or {};
table.insert(q.commandHandlers, function(msg)
	if msg == CMD_OVERHEAL then
		o[OPT_SHOWOVERHEAL] = not o[OPT_SHOWOVERHEAL];
		c:Println(q:GetText("Overheal will be displayed: ") .. q:BoolToYesNo(o[OPT_SHOWOVERHEAL]));
		return;
	elseif q:StartsWith(msg, CMD_FONTSIZE) then
		msg = msg:gsub(CMD_FONTSIZE .. " ", "");
		local fontsize = tonumber(msg);
		if fontsize and fontsize >= 10 and fontsize <= 20 then
			o[OPT_FONTSIZE] = fontsize;
			c:Println(q:GetText("Font size set to: ") .. tostring(o[OPT_FONTSIZE]));
			return;
		end
	end

	local printLines = {};
	table.insert(printLines, "[ClassicHeals] " .. q:GetText("... fontsize [number] -> sets displayed font size to given number (values between 10 and 20; default: 16)"));
	table.insert(printLines, "[ClassicHeals] " .. q:GetText("... overheal -> toggles overheal display"));
	return printLines;
end);

-- event handling
q.eventFrame = q.eventFrame or CreateFrame("Frame");
q.eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
q.eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
q.eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
q.eventFrame:HookScript("OnEvent", function(self, event, arg1)
	if event then
		if event == "ADDON_LOADED" and arg1 == "QoL_ClassicHeals" and not c.initFinished then
			c:Init();
		elseif event == "GROUP_ROSTER_UPDATE" or event == "RAID_ROSTER_UPDATE" or event == "PLAYER_ENTERING_WORLD" then
			--c:CreatePartyFrames();
		elseif event == "PLAYER_TARGET_CHANGED" then
			if targetFrame then
				UIFrameFadeOut(targetFrame, 0, 1, 0);
			end
		elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
			c:OnEvent(event, CombatLogGetCurrentEventInfo());
		end
	end
end);

-- workaround for changing raid frame interface settings
--_G["InterfaceOptionsFrameOkay"]:HookScript("OnClick", function()
--	c:ResetMemberInfo();
--end);

function c:OnEvent(event, ...)
	local timestamp, subevent, _, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = ...;

	if sourceName == q.charName and (subevent == "SPELL_HEAL" or subevent == "SPELL_PERIODIC_HEAL") then
		local spellId, spellName, spellSchool, amount, overhealing, absorbed, critical = select(12, ...);

		-- party/raid frame
		if IsInRaid() then
			local raidId, subGroupId, playerInSubGroupId = c:GetAffectedRaidMemberId(destName);
			if raidId and subGroupId and playerInSubGroupId then
				-- compactraidgroup
				if _G["CompactRaidFrameContainer"]:IsVisible() then
					if CompactRaidFrameManager_GetSetting("KeepGroupsTogether") then
						if _G["CompactRaidGroup" .. subGroupId .. "Member" .. playerInSubGroupId] then
							if not raidFrames_Pullout[raidId] then
								raidFrames_Compact[raidId] = c:CreateFrame("raidFrame_Compact" .. raidId);
							end

							c:DisplayHeal(raidFrames_Compact[raidId], _G["CompactRaidGroup" .. subGroupId .. "Member" .. playerInSubGroupId], amount, overhealing, critical);
						end
					else
						-- compactraidframe
						local compactRaidFrameId = c:GetCompactRaidFrameId(destName);
						if compactRaidFrameId and _G["CompactRaidFrame" .. compactRaidFrameId] then
							if not raidFrames_Compact[raidId] then
								raidFrames_Compact[raidId] = c:CreateFrame("raidFrame_Compact" .. raidId);
							end

							c:DisplayHeal(raidFrames_Compact[raidId], _G["CompactRaidFrame" .. compactRaidFrameId], amount, overhealing, critical);
						end
					end
				end

				-- pulloutgroup
				local pulloutGroupId = c:GetRaidPulloutId(subGroupId);
				if pulloutGroupId and _G["RaidPullout" .. pulloutGroupId .. "Button" .. playerInSubGroupId] then
					if not raidFrames_Pullout[raidId] then
						raidFrames_Pullout[raidId] = c:CreateFrame("raidFrame_Pullout" .. raidId);
					end

					c:DisplayHeal(raidFrames_Pullout[raidId], _G["RaidPullout" .. pulloutGroupId .. "Button" .. playerInSubGroupId], amount, overhealing, critical);
				end
			end
		else
			local partyMemberId = c:GetAffectedPartyMemberId(destName);
			if partyMemberId then
				if not partyFrames[partyMemberId] then
					partyFrames[partyMemberId] = c:CreateFrame("partyFrame" .. partyMemberId);
				end

				c:DisplayHeal(partyFrames[partyMemberId], _G["PartyMemberFrame" .. partyMemberId], amount, overhealing, critical);
			end
		end

		-- target frame
		if destName == GetUnitName("target") then
			if not targetFrame then
				targetFrame = c:CreateFrame("target");
			end

			targetFrame:Show();
			c:DisplayHeal(targetFrame, _G["TargetFrame"], amount, overhealing, critical);
		end
	end
end

-- main stuff
function c:DisplayHeal(frame, parentFrame, amount, overhealing, critical)
	local message;
	local framePositionSet = false;

	if overhealing > 0 then
		if o[OPT_SHOWOVERHEAL] then
			c:SetFramePosition(frame, parentFrame, true, true, critical);
			framePositionSet = true;
			message = "+" .. amount .. " (" .. overhealing .. ")";
		else
			local totalAmount = amount - overhealing;					
			if totalAmount > 0 then
				message = "+" .. totalAmount;
			else
				message = q:GetText("full");
			end
		end
	else
		message = "+" .. amount;
	end

	if not framePositionSet then
		-- set default position
		c:SetFramePosition(frame, parentFrame);
	end

	c:SetFontSize(frame, critical);
	if critical then
		UIFrameFadeIn(frame, 0.1, 0, 1);
		frame.fadeInfo.finishedFunc = function()
			UIFrameFadeOut(frame, 3, 3, 0);
		end
	else
		UIFrameFadeIn(frame, 0.3, 0, 1);
		frame.fadeInfo.finishedFunc = function()
			UIFrameFadeOut(frame, 1.5, 1.5, 0);
		end
	end

	frame:AddMessage(message, 0, 1, 0, nil, 1);
end

function c:SetFramePosition(frame, parentFrame, hasOverheal, showOverheal, critical)
	if frame:GetName() == "target" then
		frame:SetPoint("CENTER", parentFrame, 41, 2);
	else
		if IsInRaid() then
			if q:StartsWith(parentFrame:GetName(), "RaidPullout") then
				frame:SetPoint("TOP", parentFrame, 0, 2);
			else
				frame:SetPoint("TOP", parentFrame, 0, -4);
			end
		else
			frame:SetPoint("CENTER", parentFrame, -41, -2);
		end
	end
end

function c:SetFontSize(frame, critical)
	local size = o[OPT_FONTSIZE];

	if critical then
		size = size + 4;
	end

	if IsInRaid() and frame:GetName() ~= "target" then
		size = size - 4;
	end
	
	frame:SetFont("Fonts\\FRIZQT__.TTF", size, "OUTLINE");
end

--function c:CreatePartyFrames()
--	for k, _ in pairs(q:GetPartyMembers()) do
--		if not partyFrames[k] then
--			partyFrames[k] = c:CreateFrame("partyFrame" .. k);
--		end
--	end
--end

function c:GetAffectedPartyMemberId(memberName)
	for playerId, playerName in ipairs(q:GetPartyMembers()) do
		if playerName == memberName then
			return playerId;
		end
	end

	return nil;
end

-- stuff for raids
--function c:CreateRaidFrames()
--	local rm = q:GetRaidMembers();
--	for k, _ in pairs(rm[0]) do
--		if not raidFrames_Compact[k] then
--			raidFrames_Compact[k] = c:CreateFrame("raidFrame_Compact" .. k);
--		end
--		if not raidFrames_Pullout[k] then
--			raidFrames_Pullout[k] = c:CreateFrame("raidFrame_Pullout" .. k);
--		end
--	end
--end

function c:GetAffectedRaidMemberId(memberName)
	local rm = q:GetRaidMembers();
	local raidId;

	for playerId, playerName in ipairs(rm[0]) do
		if playerName == memberName then
			raidId = playerId;
			break;
		end
	end

	for subGroupId, subGroup in pairs(rm) do
		if subGroupId > 0 then
			for playerInSubGroupId, playerName in ipairs(rm[subGroupId]) do
				if playerName == memberName then
					return raidId, subGroupId, playerInSubGroupId;
				end;
			end;
		end
	end

	return nil;
end

function c:GetCompactRaidFrameId(memberName)
	for n = 1, 40 do
		local compactRaidFrameName = _G["CompactRaidFrame" .. n .. "Name"];
		if compactRaidFrameName and compactRaidFrameName:GetText() == memberName then
			return n;
		end
	end
	
	return nil;
end

function c:GetRaidPulloutId(subGroupId)
	for n = 1, 8 do
		local raidPulloutName = _G["RaidPullout" .. n .. "Name"];
		if raidPulloutName then
			local pulloutText = raidPulloutName:GetText();
			local textLen = string.len(pulloutText);
			local pulloutGroupId = tonumber(string.sub(pulloutText, textLen, textLen + 1));
			if pulloutGroupId == subGroupId then
				return n;
			end
		end
	end

	return nil;
end

-- misc stuff
function c:CreateFrame(name)
	local newFrame = CreateFrame("MessageFrame", name, UIParent);
	newFrame:SetWidth(160);
	newFrame:SetHeight(22);
	newFrame:SetFrameStrata("HIGH");
	newFrame:Show();
	return newFrame;
end

function c:Println(str)
	if str then
		q:Println(str, "ClassicHeals");
	end
end